#define _GNU_SOURCE
#include <pthread.h>
#include <unistd.h>
#include <stdio.h>
#include <stdint.h>
#include <fcntl.h>
#include <signal.h>
#include <stdlib.h>
#include <err.h>
#include <string.h>
#include <sched.h>
#include <assert.h>
#include <ucontext.h>
#include <asm/ldt.h>
#include <sys/syscall.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <sys/io.h>

#define SYSCHK(x) ({          \
  typeof(x) __res = (x);      \
  if (__res == (typeof(x))-1) \
    err(1, "SYSCHK(" #x ")"); \
  __res;                      \
})

void pin_task_to(int pid, int cpu) {
  cpu_set_t cset;
  CPU_ZERO(&cset);
  CPU_SET(cpu, &cset);
  if (sched_setaffinity(pid, sizeof(cpu_set_t), &cset))
    err(1, "affinity");
}
void pin_to(int cpu) { pin_task_to(0, cpu); }

volatile char *flip_mapping;
volatile char *mapping_both;

static void *thread_fn(void *dummy) {
  pin_to(1);

  // do PIO
  while (1) {
    volatile void *dest_addr = flip_mapping;
    unsigned long port_number = 0x08 /* status register of DMA controller */;
    asm volatile(
      // RDX: port number
      // RDI: destination address (AUTO-INC!)
      "insb\n"
    : // out
      "+D"(dest_addr)
    : // in
      "d"(port_number)
    : // clobber
      "memory", "cc"
    );
  }
  return NULL;
}

int main(void) {
  pin_to(0);
  SYSCHK(iopl(3));
  SYSCHK(mlockall(MCL_CURRENT|MCL_FUTURE));

  int mfd = SYSCHK(memfd_create("", 0));
  SYSCHK(ftruncate(mfd, 0x2000));
  mapping_both = SYSCHK(mmap(NULL, 0x2000, PROT_READ|PROT_WRITE, MAP_SHARED, mfd, 0));
  flip_mapping = SYSCHK(mmap(NULL, 0x1000, PROT_READ|PROT_WRITE, MAP_SHARED, mfd, 0));

  pthread_t thread;
  if (pthread_create(&thread, NULL, thread_fn, NULL))
    errx(1, "pthread_create");

  while (1) {
    // make first page visible
    SYSCHK(mmap((void*)flip_mapping, 0x1000, PROT_READ|PROT_WRITE, MAP_SHARED|MAP_FIXED, mfd, 0));
    // wait for write to start on first page
    for (int i=0; i<100000; i++) /*spin*/;
    // replace page with second and do TLB flush
    SYSCHK(mmap((void*)flip_mapping, 0x1000, PROT_READ|PROT_WRITE, MAP_SHARED|MAP_FIXED, mfd, 0x1000));
    // first page should now no longer be accessible by thread. overwrite its contents...
    mapping_both[0] = 0xff;
    // ... spin a bit for the IO to complete ...
    for (int i=0; i<100000; i++) /*spin*/;
    // ... and see if it was illegally clobbered?
    unsigned char cur_value = mapping_both[0];
    if (cur_value != 0xff) {
      printf("illegally clobbered from ff to %02hhx\n", cur_value);
    }
  }
}


